"""An example showing how to run containerized NetLogo on HPC system."""

import os
from kunefe import Kunefe

# define the constants
USERNAME = "JaneDoe"
# folders
OUTPUT_FOLDER_LOCAL = "./output/"    # local folder to save output
REMOTE_HOME = f"/home/{USERNAME}/"   # home folder on the remote system
OUTPUT_FOLDER_REMOTE = REMOTE_HOME + "test_folder/"
# specify the Docker image
NETLOGO_VERSION = "6.3.0"
DOCKER_IMAGE = f"comses/netlogo:{NETLOGO_VERSION}"
# specify the apptainer image
SIF_FILE_NAME = f"netlogo_{NETLOGO_VERSION}.sif"
SIF_FILE_PATH_LOCAL = OUTPUT_FOLDER_LOCAL + SIF_FILE_NAME
SIF_FILE_PATH_REMOTE = OUTPUT_FOLDER_REMOTE + SIF_FILE_NAME

# We will run the NetLogo simulation on Snellius cluster:
# https://www.surf.nl/diensten/snellius-de-nationale-supercomputer
kunefe = Kunefe(username=USERNAME, hostname="snellius.surf.nl", port=22)

# create a local folder to save the generated files and the output
try:
    os.mkdir(OUTPUT_FOLDER_LOCAL)
    print(f"Directory '{OUTPUT_FOLDER_LOCAL}' created.")
except FileExistsError:
    print(f"Directory '{OUTPUT_FOLDER_LOCAL}' already exists.")

# build apptainer image from a Docker image
kunefe.build_apptainer_image(
    docker_image=DOCKER_IMAGE, sif_file_name=SIF_FILE_PATH_LOCAL
)

# make a connection to the remote
kunefe.connect_remote()

# create a folder which the files will be copied to
kunefe.create_remote_folder(remote_folder=OUTPUT_FOLDER_REMOTE)

# define the job name and the command to be used to run NetLogo
JOB_NAME = "kunefe_netlogo_experiment_job"
NETLOGO_COMMAND = """/opt/netlogo/netlogo-headless.sh \
--model 'model_path' \
--experiment 'experiment_name' \
--table 'table_name'
"""

# generate a job script to be submitted to HPC job queue
kunefe.generate_job_script(
    job_name=JOB_NAME,
    sif_file_path=SIF_FILE_PATH_REMOTE,
    command=NETLOGO_COMMAND,
    env_vars="JAVA_TOOL_OPTIONS=-Xmx8G",
    job_file_path=OUTPUT_FOLDER_LOCAL,
    job_time="0:1:00",
    template_name="generic",
)

# copy the required files and the job script to the remote system
kunefe.put_files(remote_folder=OUTPUT_FOLDER_REMOTE, local_folder=OUTPUT_FOLDER_LOCAL, verbose=True)

# submit the job
job_id, stdin, stdout, stderr = kunefe.submit_job(
    job_file=f"{OUTPUT_FOLDER_REMOTE}/{JOB_NAME}.sh"
)

# monitor the queue
kunefe.watch_slurm_queue()

# the method below can be used to retrieve the job results
# kunefe.get_files(remote_folder=OUTPUT_FOLDER_REMOTE, local_folder=f"{OUTPUT_FOLDER_LOCAL}/result")



"""Tests remote functions of kunefe ."""
from time import sleep
import docker
import pytest
from kunefe import Kunefe


# To see the print messages run with '--capture=no':
# pytest --capture=no -v tests/test_kunefe.py
@pytest.fixture(scope='session', autouse=True)
def slurm_service():
    """Start the SLURM service before running the tests."""
    container_name = "test_kunefe"
    docker_image_name = "xenonmiddleware/slurm:latest"

    # https://docker-py.readthedocs.io/en/stable/containers.html
    # start the SLURM service before tests
    client = docker.from_env()
    print('\nStarting the Docker container')
    container = client.containers.run(
        name=container_name,
        image=docker_image_name,
        detach=True,
        ports={'22/tcp': 10022}
    )
    sleep(10)
    container_state = container.attrs['State']
    assert container_state['Status'] == 'created'

    yield container

    # stop the SLURM service after the last test
    print('\nStopping the Docker container')
    container.stop()
    container.remove(force=True)


# https://docs.pytest.org/en/latest/how-to/monkeypatch.html
# to only test this function run:
# pytest --capture=no -v tests/test_kunefe.py::test_connection_with_password
def test_connection_with_password(monkeypatch, slurm_service):
    """Test if kunefe can ssh to the SLURM Docker container using a password."""
    responses = iter(['javagat'])  # the default password
    monkeypatch.setattr('getpass.getpass', lambda _: next(responses))
    kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
    kunefe.connect_remote()
    # assert Kunefe(username="xenon", hostname="localhost", port=10022) == ("Me", "xxx")


# def test_create_remote_folder():
#     """Test create_remote_folder
#     """
#     pass


# def test_get_files():
#     """Test get_files
#     """
#     pass


# def test_put_files():
#     """Test put_files
#     """
#     pass


# def test_submit_job():
#     """Test submit_job
#     """
#     pass


def test_run_remote_command(monkeypatch, slurm_service):
    """Test if a command runs succesfully on the remote system."""
    responses = iter(['javagat'])  # the default password
    monkeypatch.setattr('getpass.getpass', lambda _: next(responses))
    kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
    kunefe.connect_remote()
    _, stdout, stderr = kunefe.run_remote_command(command='ls /home/xenon')
    assert stderr == '', 'there was an error'
    assert stdout == 'filesystem-test-fixture\ntest-slurm.job\n', 'the command output does not match'


# def test_watch_slurm_queue():
#     """Test watch_slurm_queue
#     """
#     pass



"""Tests local functions of kunefe."""
import os
import pytest
from kunefe import Kunefe

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"

@pytest.fixture
def generic_job_sample_file():
    """Test Employee Fixture."""
    with open("tests/data/kunefe_generic_job.sh", mode="r", encoding="utf-8") as file:
        data = file.read().rstrip()
        return data


#@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Skipping the test: Test does not work in Github Actions.")
def test_build_apptainer_image(tmp_path):
    """Test build_apptainer_image."""
    print(f'\nworkdir: {tmp_path}')
    kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
    assert kunefe.build_apptainer_image(
        docker_image='comses/netlogo:6.3.0',
        sif_file_name=f'{tmp_path}/netlogo_6.3.0.sif'
    )


#@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Skipping the test: Test does not work in Github Actions.")
def test_build_apptainer_image_fail(tmp_path):
    """Test build_apptainer_image."""
    print(f'\nworkdir: {tmp_path}')
    kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
    assert not kunefe.build_apptainer_image(
        docker_image='someorg/someimage:99.99.0',
        sif_file_name=f'{tmp_path}/someimage_99.99.0.sif'
    )


# # See https://stackoverflow.com/a/52877169
# def test_compare_generated_apptainer_image():
#     """Test build_apptainer_image
#     """
#     pass


def test_check_local_command_exists():
    """Test check_local_command_exists."""
    kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
    assert kunefe.check_local_command_exists(command='ls')


def test_check_local_command_exists_fail():
    """Test check_local_command_exists."""
    kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
    assert not kunefe.check_local_command_exists(command='thisdoesnotexist')


# to only test this function run:
# pytest --capture=no -v tests/test_kunefe.py::test_check_required_tools
def test_check_required_tools():
    """Test if required tools can be found."""
    kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
    requirements_status = kunefe.check_required_tools(['docker', 'rsync'])
    print(f'Requirements status: {requirements_status}')
    assert requirements_status is True, 'all the required tools need to be installed on your system'


# to only test this function run:
# pytest --capture=no -v tests/test_kunefe.py::test_check_required_tools_fail
def test_check_required_tools_fail():
    """Test if check_required_tools fails for a command that does not exist."""
    kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
    requirements_status = kunefe.check_required_tools(['docker', 'rsync', 'bsxcommand'])
    print(f'Requirements status: {requirements_status}')
    assert requirements_status is False, 'should fail for bsxcommand'


# https://docs.pytest.org/en/7.1.x/how-to/tmp_path.html
def test_generate_job_script(tmp_path, generic_job_sample_file):
    """Test generate_job_script."""
    print(f'\nworkdir: {tmp_path}')
    # print(generic_job_sample_file)

    kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
    kunefe.generate_job_script(
        job_name='pytest_job',
        sif_file_path="/home/xenon/myapp_0.1.0.sif",
        command="ls /home/xenon",
        env_vars="PATH=$PATH:/home/xenon",
        job_time='0:30:00',
        job_file_path=tmp_path
    )

    generated_file_content = None
    generated_file_path = f"{tmp_path}/pytest_job.sh"
    with open(generated_file_path, mode="r", encoding="utf-8") as file:
        generated_file_content = file.read().rstrip()
    # print(generated_file_content)

    assert generic_job_sample_file == generated_file_content



"""Documentation about kunefe."""
import logging
from .kunefe import Kunefe

logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Faruk D."
__email__ = "f.diblen@esciencecenter.nl"
__version__ = "0.5.0"

__all__ = [
    "Kunefe"
]



"""Tests for the kunefe."""
from time import sleep
import docker


def test_docker_start():
    """Check if Docker container runs."""
    container_name = "test_docker_start"
    docker_image_name = "xenonmiddleware/slurm:latest"

    # https://docker-py.readthedocs.io/en/stable/containers.html
    client = docker.from_env()
    container = client.containers.run(
        name=container_name,
        image=docker_image_name,
        detach=True,
        ports={'22/tcp': 10022}
    )

    # timeout = 20
    # stop_time = 1
    # elapsed_time = 0
    # while container.status != 'running' and elapsed_time < timeout:
    #     sleep(stop_time)
    #     elapsed_time += stop_time
    #     continue

    sleep(10)
    container_state = container.attrs['State']
    # container.wait()
    container.stop()
    container.remove(force=True)

    assert container_state['Status'] == 'created'
    # assert container.status == ['created', 'running'], 'container is not created or running'


def test_slurm_queue():
    """Check if the SLURM service of Xenon Docker image works."""
    container_name = "test_slurm_queue"
    docker_image_name = "xenonmiddleware/slurm:latest"

    # https://docker-py.readthedocs.io/en/stable/containers.html
    client = docker.from_env()
    container = client.containers.run(
        name=container_name,
        image=docker_image_name,
        detach=True,
        ports={'22/tcp': 10022}
    )

    sleep(240)
    exit_code, output = container.exec_run(
        'squeue --all',
        stderr=True,
        stdout=True
    )

    container.stop()
    container.remove(force=True)

    assert exit_code == 0, output



"""Documentation about the kunefe module.

Returns:
    _type_: _description_
"""

import atexit
import getpass
import os
import subprocess
import sys
import time
from shutil import which
from stat import S_ISDIR
from stat import S_ISREG
from typing import Tuple
import jinja2
import paramiko

# setup logging
DEFAULT_LOG_FILENAME = "kunefe.log"
paramiko.util.log_to_file(DEFAULT_LOG_FILENAME)


class Kunefe:
    """Submit jobs to SLURM cluster.

    Attributes:
        username (str): login name of the user.
        hostname (str): hostname or the IP address of the remote system.
        port (int): SSH port to be used by the clients.
        password (int): password to connect to the remote.
        ssh_client: ssh client to connect and run commands on the remote system.
        sftp_client: sftp client to copy files from and to a remote system.
    """

    def __init__(self, username: str, hostname: str, port: int) -> None:
        """Initialize Kunefe class with username, hostname and port.

        Args:
            username (str): login name of the user.
            hostname (str): hostname or the IP address of the remote system.
            port (int): SSH port to be used by the clients.

        Returns:
            None
        """
        self.username = username
        self.hostname = hostname
        self.port = port
        self.password: str
        self.ssh_client: paramiko.SSHClient
        self.sftp_client: paramiko.SFTPClient
        atexit.register(self.cleanup)
        return None

    def set_password(self) -> str:
        """Sets user password. The password is not echoed.

        Returns:
            password (str): password of the user
        """
        password = getpass.getpass(
            f"password for {self.username}@{self.hostname}: "
        )
        return password

    def set_ssh_client(self) -> paramiko.SSHClient:
        """Creates an SSH client and connects to remote system.

        Returns:
            ssh_client (paramiko.SSHClient): ssh client
        """
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        return ssh_client

    def connect_remote(self) -> None:
        """Creates an ssh and sftp clients, prompts for user password and connects to the remote host.

        Returns:
            None
        """
        self.ssh_client = self.set_ssh_client()
        self.password = self.set_password()
        self.ssh_client.connect(
            hostname=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
        )
        self.sftp_client = self.ssh_client.open_sftp()
        return None

    def create_remote_folder(self, remote_folder: str) -> None:
        """Create a folder in the remote system.

        Args:
            remote_folder (str): path of the folder to be created on the remote system.

        Returns:
            None
        """
        try:
            self.sftp_client.mkdir(remote_folder)
        except IOError:
            print(f"(assuming {remote_folder}/ already exists)")
        return None

    def get_files(self, remote_folder: str, local_folder: str = "./") -> None:
        """Get files from the remote system.

        Args:
            remote_folder (str): path of the remote folder to copy the files from.
            local_folder (str, optional): path of the host folder to copy the files to. Defaults to "./".

        Returns:
            None
        """
        if not os.path.exists(local_folder):
            os.mkdir(local_folder)

        for entry in self.sftp_client.listdir_attr(remote_folder):
            remote_path = remote_folder + "/" + entry.filename
            local_path = os.path.join(local_folder, entry.filename)
            mode = entry.st_mode
            if mode is None:
                # TODO: throw an exception here
                print("Path type is None")
            else:
                if S_ISDIR(mode):
                    try:
                        os.mkdir(local_path)
                    except OSError:
                        pass
                    self.get_files(remote_path, local_path)
                elif S_ISREG(mode):
                    self.sftp_client.get(remote_path, local_path)
                else:
                    # TODO: throw an exception here
                    print("Unknown type encountered when running get_files method")
        return None

    def put_files(
        self,
        remote_folder: str = "~",
        local_folder: str = "./",
        verbose: bool = False,
    ) -> None:
        """Copy files to the remote system.

        Args:
            remote_folder (str): path on a remote system to copy the files to.
            local_folder (str): path to copy the files from.
            verbose (bool): show verbose info when copying.

        Returns:
            None
        """
        if remote_folder == "~":
            remote_folder = os.path.expanduser("~")

        local_folder_dirname = os.path.dirname(local_folder)
        local_folder_basename = os.path.basename(local_folder)

        if verbose:
            print(f"local_folder_dirname: {local_folder_dirname}")
            print(f"local_folder_basename: {local_folder_basename}")

        for dirpath, dirnames, filenames in os.walk(local_folder):
            file_dirpath = dirpath[len(local_folder) + 1:]
            remote_path = os.path.join(
                remote_folder, local_folder_basename, file_dirpath
            )
            if verbose:
                print("\n" + "-" * 20 + "\n")
                print(f"dirpath: {dirpath}")
                print(f"dirnames: {dirnames}")
                print(f"remote_path: {remote_path}")
                print(f"file_dirpath: {file_dirpath}")

            try:
                self.sftp_client.listdir(remote_path)
            except IOError:
                self.sftp_client.mkdir(remote_path)

            for filename in filenames:
                from_path = os.path.join(dirpath, filename)
                to_path = os.path.join(remote_path, filename)
                self.sftp_client.put(from_path, to_path)
                if verbose:
                    print(
                        f"\ncopying: {filename}\n  from: {from_path}\n  to: {to_path}"
                    )
        return None

    def submit_job(self, job_file: str, verbose: bool = False) -> Tuple[int, str, str, str]:
        """Submit job to SLURM cluster.

        Args:
            job_file (str): full path of the job script to be submitted.
            verbose (bool): run in verbose mode.

        Returns:
            tuple(int, str, str, str): job_id, stdin, stdout, stderr
        """
        stdin, stdout, stderr = self.ssh_client.exec_command(
            f"sbatch {job_file}"
        )
        job_id = int(stdout.read().decode().split()[-1])
        if verbose:
            print(f"Submitted job with id: {job_id}")
        return (job_id, stdin.read().decode(), stdout.read().decode(), stderr.read().decode())

    def build_apptainer_image(self, docker_image: str, sif_file_name: str = 'app.sif') -> bool:
        """Builds an Apptainer image from a Docker image.

        Args:
            docker_image (str): docker image name to be used to build an apptainer image.
            sif_file_name (str, optional): name of the apptainer image (sif) to be built. Defaults to 'app.sif'.

        Returns:
            bool: True if the apptainer image was successfully built. Otherwise, returns False.
        """
        build_command = f"apptainer pull {sif_file_name} docker://{docker_image}"

        process = subprocess.Popen(build_command, shell=True)
        process.wait()

        # TODO: handle process status mor carefully: what happens if the file exists?
        if os.path.isfile(sif_file_name):
            print(f"Generated {sif_file_name}")
            # print(process.stdout)
            return True
        else:
            print("Image generation has failed.")
            print(process.stderr)
            return False

    def install_apptainer_on_remote(self, install_path: str = "~") -> bool:
        """Installs Apptainer on a remote system in unprivileged mode.

        Args:
            install_path (str): path to install apptainer binary.

        Returns:
            bool: True if the apptainer installation was successful. Otherwise, returns False.
        """
        if install_path == "~":
            install_path = os.path.expanduser("~")

        print(f"installing apptainer at {install_path}")

        install_script_url = "https://raw.githubusercontent.com/apptainer/apptainer/main/tools/install-unprivileged.sh"
        exe_path = f"{install_path}/bin/apptainer"

        install_command = f"curl -s {install_script_url} | bash -s - {install_path}"
        # print(f"running: {install_command}")
        _, _, stderr_install = self.run_remote_command(
            command=install_command, timeout=30, flush=False, show_stdout=True)

        check_install_command = f"file {exe_path}"
        # print(f"running: {check_install_command}")
        _, _, stderr_check = self.run_remote_command(
            command=check_install_command, timeout=30, flush=False, show_stdout=True)

        # TODO: add the exectable to $PATH and check the executable on the remote system: command -v apptainer
        if stderr_install == "" or stderr_check == "":
            print(f"Installed at {exe_path}")
            return True
        else:
            print("Apptainer installation has failed.")
            print(f"installation error: {stderr_install}")
            print(f"installation check error:{stderr_check}")
            return False

    def check_local_command_exists(self, command: str) -> bool:
        """Check whether `command` is on PATH and marked as executable.

        Args:
            command (str): a command to be checked.

        Returns:
            bool: True if command exists. Otherwise False.

        Examples:
            >>> from kunefe import Kunefe
            >>> kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
            >>> kunefe.check_local_command_exists(command='ls')
            True
            >>> kunefe.check_local_command_exists(command='thisdoesnotexist')
            False
        """
        return which(command) is not None

    def check_required_tools(self, command_list: list[str]) -> bool:
        """Check whether all required commands are available.

        Args:
            command_list (list[str]): a list of tools to be checked.

        Returns:
            bool: True if all the tools exist. Otherwise False.

        Examples:
            >>> from kunefe import Kunefe
            >>> kunefe = Kunefe(username="xenon", hostname="localhost", port=10022)
            >>> kunefe.check_required_tools(['ls', 'rsync', 'cp'])
            True
            >>> kunefe.check_required_tools(['docker', 'rsync', 'bsxcommand'])
            False
        """
        # TODO: also check the required versions
        if all(self.check_local_command_exists(command) for command in command_list):
            return True
        else:
            return False

    def generate_job_script(
        self,
        job_name: str,
        sif_file_path: str,
        command: str,
        env_vars: str,
        job_time: str,
        job_file_path: str = './',
        template_name: str = 'generic'
    ) -> None:
        """Generate a batch script file for job submission.

        Args:
            job_name (str): name of the job to be used when submitting.
            sif_file_path (str): path of the Apptainer image.
            command (str): a command to be executed using Apptainer image.
            env_vars (str): environment variables to be used when submitting the job.
            job_time (str): time limit for the job.
            job_file_path (str, optional): path to save the generated batch job script. Defaults to './'.
            template_name (str, optional): name of the template to be used. Defaults to 'generic'.

        Returns:
            None
        """
        parent_dir = os.path.dirname(__file__)
        templates_folder = os.path.join(parent_dir, "templates")
        environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(templates_folder)
        )
        template = environment.get_template(f"{template_name}_job.jinja")

        filename = f"{job_file_path}/{job_name}.sh"
        content = template.render(
            job_name=job_name,
            sif_file_path=sif_file_path,
            job_time=job_time,
            command=command,
            env_vars=env_vars
        )

        with open(filename, mode="w", encoding="utf-8") as message:
            message.write(content)
            print(f"Batch job file was saved as {filename}")
        return None

    def run_remote_command(
        self, command: str, timeout: int = 5, flush: bool = False, show_stdout: bool = False
    ) -> list[str]:
        """Run a command on a remote system.

        Args:
            command (str): command to be executed on the remote system.
            timeout (int, optional): time to wait before considering the command as failed. Defaults to 5.
            flush (bool, optional): flush the output. Defaults to False.
            show_stdout (bool, optional): prints the stdout. Defaults to False.

        Returns:
            list[str]: stdin, stdout, stderr
        """
        self.ssh_client.invoke_shell()
        stdin, stdout, stderr = self.ssh_client.exec_command(
            command=command,
            timeout=timeout
        )

        if show_stdout:
            stdout_lines = stdout.readlines()
            for line in stdout_lines:
                if flush:
                    print(line, end="", flush=True)
                else:
                    print(line, end="")

            # https://stackoverflow.com/a/11474509
            if flush:
                sys.stdout.write(
                    "\033[F" * len(stdout_lines)
                )  # Cursor up for X number of lines

        return [stdin.read().decode('ascii') if stdin.readable() else '',
                stdout.read().decode('ascii') if stdout.readable() else '',
                stderr.read().decode('ascii') if stderr.readable() else '']

    def watch_slurm_queue(self, sleep_time: float = 5.0) -> None:  # pragma: no cover
        """Watches the SLURM job queue.

        Args:
            sleep_time (float, optional): time to wait before refreshing the queue status. Defaults to 5.0.

        Returns:
            None
        """
        command = 'squeue --all'
        while True:
            _, _, stderr = self.run_remote_command(command=command, timeout=5, flush=True, show_stdout=True)
            if stderr != '':
                break
            time.sleep(sleep_time)
        return None

    def cleanup(self) -> None:  # pragma: no cover
        """Destructor method to clean things up.

        Returns:
            None
        """
        print("Running cleanup...")
        return None
